/*----------------------------------------------------------------------
    Document Image Labeling                   dlabel  ver1.31
        Written by  H.Goto, Mar 1999
        Modified by H.Goto, Jun 1999
        Modified by H.Goto, Sep 1999
        Modified by H.Goto, Oct 1999
        Modified by H.Goto, Nov 1999
        Modified by H.Goto, Jul 2000
        Modified by H.Goto, Jul 2002
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1999-2002  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

//#define		SaveFilesLimit	20	/* for debug use only */

#define		FastXlatMode
#define		Merge5x5Mode

#define		Min_ObjectSize	16	/* not so sensitive */

#define		SUBREGIONSIZE	50	/* must be a multiple of 2 */
#define		WINDOWSIZE	60
#define		WINDOWOFFSET	-((WINDOWSIZE - SUBREGIONSIZE) /2)

#define		AllowConflict		4
#define		AllowConflictEdgeRatio	25
#define		IgnoreFragmentSize	8

#define		TruncateTinyGroups
#define		TruncLevel	16
#define		TruncCounts	4

#define		IMGBUFDEPTH	8

#define		GrayLevels	256
#define		MAXPEAKS	(GrayLevels /4)
#define		Max_mwidth	0x1000

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>

#include	"utypes.h"

#include	"CRect.h"
#include	"objgrp.h"
#include	"ufilep.h"
#include	"siplib.h"
#include	"xiplib.h"
#include	"sgplib.h"
#include	"iset.h"
#include	"labelset.h"
#include	"gettime.h"
#include	"findcol_com.h"
#include	"findgray_pd.h"
#include	"edgecomp.h"


		/* minimum difference at peak merging */
static int	mindiff[] = { 4,8,16,24,32,48,64,-1 };

static int	sw_verbose = 0;
static int	sw_weighted = 0;
static int	sw_smooth = 0;
static int	sw_edgecomp = -1;
static int	sw_decomp = 0;

static double	utime_load,stime_load;
static double	utime_filt,stime_filt;
static double	utime_edge,stime_edge;
static double	utime_label,stime_label;
static double	utime_merge,stime_merge;
static double	utime_reduce,stime_reduce;
static double	utime_mklab,stime_mklab;
static double	utime_decomp,stime_decomp;


//typedef unsigned short	GLAB;
typedef unsigned char	GLAB;

typedef struct {
	int		ncolor;
	CCluster	collist[MAXPEAKS];
	int		glabel[MAXPEAKS];
	char		**inh_table;
} MTCELL;




/*----------------------------------------------
    Find local gray levels
----------------------------------------------*/

static void find_graylevels(SIPImage *image,MTCELL **cmatrix, \
		int mwidth,int mheight,int weighted){
	int	mx,my;
	int	n_cluster;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			n_cluster = findgray_apd(image, \
				SUBREGIONSIZE * mx + WINDOWOFFSET, \
				SUBREGIONSIZE * my + WINDOWOFFSET, \
				WINDOWSIZE,WINDOWSIZE, \
				cmatrix[my][mx].collist,4,-1,weighted);
			if ( n_cluster < 0 )  n_cluster = 0;
			cmatrix[my][mx].ncolor = n_cluster;
		}
	}
	return;
}




/*----------------------------------------------
    Make gray-to-label translation table
----------------------------------------------*/

static int make_xlat(MTCELL *cmatp,GLAB *xlat_lab){
	int	i,j;
	int	divpos[MAXPEAKS +1];
	divpos[0] = 0;
	for ( i=0 ; i < cmatp->ncolor -1 ; i++ ){
		divpos[i+1] = (cmatp->collist[i+1].g \
				+ cmatp->collist[i].g) /2;
	}
	divpos[cmatp->ncolor] = GrayLevels;

	for ( j=0 ; j<cmatp->ncolor ; j++ ){
		for ( i=divpos[j] ; i<divpos[j+1] ; i++ ){
			xlat_lab[i] = j;
		}
	}
	return(0);
}




/*----------------------------------------------
    Find close gray id.
----------------------------------------------*/

static int find_closegray(MTCELL *cmatp,int gray){
	int	i;
	int	mindist,minpos,dist;
	mindist = 0x7fffffff;  minpos = -1;
	for ( i=0 ; i < cmatp->ncolor ; i++ ){
		dist = abs(gray - cmatp->collist[i].g);
		if ( mindist > dist ){
			mindist = dist;  minpos = i;
		}
	}
	return(minpos);
}




/*----------------------------------------------
    Make reduced image
----------------------------------------------*/

static int make_reducedimage(SIPImage *labelimage,SIPImage *rimage, \
		MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my,x,y;
	GLAB	*src_lab;
	uchar	*dst;
	MTCELL	*cmatp;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];

			for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
				src_lab = (GLAB *)sip_getimgptr(labelimage, \
					y + my * SUBREGIONSIZE);
				src_lab += mx * SUBREGIONSIZE;
				dst = (uchar *)sip_getimgptr(rimage, \
					y + my * SUBREGIONSIZE);
				dst += mx * SUBREGIONSIZE;
				for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
					dst[x] = (uchar)cmatp->collist[ src_lab[x] ].g;
				}
			}
		}
	}
	return(0);
}




/*------------------------------------------------------
	Noise Reduction (size filter)
------------------------------------------------------*/

static int accumulate_neighbors(SIPImage *image,int x0,int y0,int *hist){
	int	x,y;
	int	xs,xe,ys,ye;
	uchar	d,*p;
	if ( 0 > (xs = x0 -1) )  xs = 0;
	if ( image->width <= (xe = x0 +1) )  xe = image->width -1;
	if ( 0 > (ys = y0 -1) )  ys = 0;
	if ( image->height <= (ye = y0 +1) )  ye = image->height -1;
	d = ((uchar *)sip_getimgptr(image,y0))[x0];
	for ( y=ys ; y<=ye ; y++ ){
		p = (uchar *)sip_getimgptr(image,y);
		for ( x=xs ; x<=xe ; x++ ){
			if ( p[x] != d )  hist[ (uint)p[x] ]++;
		}
	}
	return(0);
}


static int find_surround(GRPLST *ccpool,int cid,CRects *runpool, \
		SIPImage *labelimage,int *neiblist,int *nneib,int maxneib){
	int	i,x,nrid,n;
	CRect	*rect;
	uchar	*p;
	int	*hist;
	int	maxcount,maxpos;
	nrid = ccpool->getidcounts(cid);
	if ( 0 == (hist = new int[ ccpool->groups > maxneib ? ccpool->groups : maxneib ]) )  return(-1);
	for ( i=0 ; i < ccpool->groups ; i++ )  hist[i] = 0;
	for ( i=0 ; i<nrid ; i++ ){
		rect = runpool->getrect( ccpool->getidlist(cid)[i] );
		p = (uchar *)sip_getimgptr(labelimage,rect->y1);
		for ( x = rect->x1 ; x <= rect->x2 ; x++ ){
			accumulate_neighbors(labelimage,x,rect->y1,hist);
		}
	}

	/* find 1st through maxneib-th populated label */
	for ( n=*nneib=0 ; n < ( ccpool->groups < maxneib ? ccpool->groups : maxneib ) ; n++ ){
		maxpos = -1;  maxcount = 0;
		for ( i=0 ; i < ccpool->groups ; i++ ){
			if ( hist[i] <= maxcount )  continue;
			maxcount = hist[i];
			maxpos = i;
		}
		if ( maxpos == -1 )  break;
		hist[ maxpos ] = 0;
		neiblist[ (*nneib)++ ] = maxpos;
	}
	delete []hist;
	return( *nneib );
}


static int NR_size(SIPImage *image,SIPImage *labelimage,int minsize){
	int	cid,nrid,npixel;
	int	x,i;
	CRect	*rect;
	uchar	*lp;
	GRPLST	ccpool;
//	CRects	runpool(4096);
	CRects	runpool(1024);
	int	neiblist[8];
	int	nneib;
	/* create 8-connected components */
	xip_CreateCC(labelimage,&ccpool,&runpool,1);
	for ( cid=0 ; cid < ccpool.groups ; cid++ ){
		if ( 0 >= (nrid = ccpool.getidcounts(cid)) )  continue;
		for ( i=npixel=0 ; i<nrid ; i++ ){
			npixel += runpool.getrect( ccpool.getidlist(cid)[i] )->x2;
			npixel -= runpool.getrect( ccpool.getidlist(cid)[i] )->x1;
			npixel++;
		}
		if ( npixel >= minsize )  continue;

		if ( 0 >= find_surround(&ccpool,cid,&runpool, \
			labelimage,neiblist,&nneib,8) )  continue;
		for ( i=0 ; i<nrid ; i++ ){
			rect = runpool.getrect( ccpool.getidlist(cid)[i] );
			lp = (uchar *)sip_getimgptr(labelimage,rect->y1);
			for ( x = rect->x1 ; x <= rect->x2 ; x++ ){
				lp[x] = (uchar)neiblist[0];
			}
		}
	}
	ccpool.clearall();
	runpool.clear();
	return(0);
}




/*----------------------------------------------
    Make local-label image (for grayscale)
----------------------------------------------*/

static int make_labelimage_loc(SIPImage *image,SIPImage *labelimage, \
		MTCELL **cmatrix,int mx,int my,int colreduce){
	int	x,y,i,ncol;
	uchar	*src,*dst;
	GLAB	*dst_lab;
	int	downconv = 0;
#ifdef	FastXlatMode
	GLAB	xlat_lab[GrayLevels];
#endif
	MTCELL	*cmatp;
	int	hist[MAXPEAKS];
	SIPImage	*tmpimage,*rimage,*label_local;
	SIPRectangle	srect = { 0,0,SUBREGIONSIZE,SUBREGIONSIZE };
	SIPRectangle	drect = { 0,0,SUBREGIONSIZE,SUBREGIONSIZE };
	if ( 0 == (tmpimage = sip_CreateImage(SUBREGIONSIZE,SUBREGIONSIZE,8)) ){
		return(-1);
	}
	if ( 0 == (rimage = sip_CreateImage(SUBREGIONSIZE,SUBREGIONSIZE,8)) ){
		sip_DestroyImage(tmpimage);
		return(-1);
	}
	if ( 0 == (label_local = sip_CreateImage(SUBREGIONSIZE,SUBREGIONSIZE,8)) ){
		sip_DestroyImage(rimage);
		sip_DestroyImage(tmpimage);
		return(-1);
	}
	cmatp = &cmatrix[my][mx];
#ifdef FastXlatMode
	make_xlat(cmatp,xlat_lab);
#endif

	srect.x = mx * SUBREGIONSIZE;
	srect.y = my * SUBREGIONSIZE;
	sip_CopyArea(image,&srect,tmpimage,&drect);

	for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
		src = (uchar *)sip_getimgptr(tmpimage,y);
		dst = (uchar *)sip_getimgptr(rimage,y);
		for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
#ifndef FastXlatMode
			int	minpos;
			minpos = find_closegray(cmatp,\
					(int)((uint)src[x]));
			dst[x] = (uchar)cmatp->collist[minpos].g;
#else
			dst[x] = (uchar)cmatp->collist[ xlat_lab[ ((uint)src[x]) ] ].g;
#endif	/* FastXlatMode */
		}
	}

	/* edge compensation */
	if ( cmatp->ncolor > 2 && sw_edgecomp ){
		compensate_edge(rimage,tmpimage,rimage,2,8,6);
	}

	for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
		src = (uchar *)sip_getimgptr(rimage,y);
		dst = (uchar *)sip_getimgptr(label_local,y);
		for ( x=0 ; x<SUBREGIONSIZE ; x+=2 ){
#ifndef FastXlatMode
			dst[x  ] = (uchar)find_closegray(cmatp,\
					(int)((uint)src[x  ]));
			dst[x+1] = (uchar)find_closegray(cmatp,\
					(int)((uint)src[x+1]));
#else
			dst[x  ] = (uchar)xlat_lab[ (int)((uint)src[x  ]) ];
			dst[x+1] = (uchar)xlat_lab[ (int)((uint)src[x+1]) ];
#endif	/* FastXlatMode */
		}
	}

	/* filter out tiny blobs */
	if ( cmatp->ncolor != 1 ){
		NR_size(label_local,label_local,4);
	}

	/* reduce number of gray levels */
	ncol = cmatp->ncolor;
	if ( colreduce && ncol > 2 ){
		for ( i=0 ; i<ncol ; i++ ){
			hist[i] = 0;
		}
		for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
			src = (uchar *)sip_getimgptr(label_local,y);
			for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
				hist[ (uint)src[x  ] ]++;
			}
		}
		for ( i=cmatp->ncolor=0 ; i<ncol ; i++ ){
			cmatp->collist[cmatp->ncolor] = cmatp->collist[i];
			if ( 4 <= hist[i] )  cmatp->ncolor++;
		}
	}
	if ( cmatp->ncolor != ncol ){
		downconv = 1;

		/* label it once again */
		if ( 0 > make_labelimage_loc(image,labelimage,cmatrix,mx,my,0) ){
			downconv = -1;
		}
	}
	else{
		/* copy temporary label image to real label image */
		for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
			src = (uchar *)sip_getimgptr(label_local,y);
			dst_lab = (GLAB *)sip_getimgptr(labelimage, \
				y + my * SUBREGIONSIZE);
			dst_lab += mx * SUBREGIONSIZE;
			for ( x=0 ; x<SUBREGIONSIZE ; x+=2 ){
				dst_lab[x  ] = (GLAB)src[x  ];
				dst_lab[x+1] = (GLAB)src[x+1];
			}
		}
	}

	sip_DestroyImage(label_local);
	sip_DestroyImage(rimage);
	sip_DestroyImage(tmpimage);

	return(downconv);
}




/*----------------------------------------------
    Make all local label images
----------------------------------------------*/

static int make_locallabelimage(SIPImage *image,SIPImage *labelimage, \
		MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			make_labelimage_loc(image,labelimage,cmatrix,mx,my,1);
		}
	}
	return(0);
}




/*------------------------------------------------------
	Inhibition Table Memory Management
------------------------------------------------------*/

static char ** alloc_inh_table(int size){
	char	**table;
	int	i;
	if ( size <= 0 )  size = 1;
	if ( 0 == (table = new char * [size]) )  return(0);
	for ( i=0 ; i<size ; i++ )  table[i] = 0;
	for ( i=0 ; i<size ; i++ ){
		if ( 0 == (table[i] = new char[size]) ){
			i--;
			for ( ; i >= 0 ; i-- )  delete []table[i];
			delete []table;
			return(0);
		}
	}
	return(table);
}


static int dealloc_inh_table(char **table,int size){
	int	i;
	if ( table == 0 )  return(0);
	for ( i=0 ; i<size ; i++ ){
		if ( table[i] )  delete []table[i];
	}
	delete []table;
	return(0);
}




/*------------------------------------------------------
	Make Connection-Inhibition Table
------------------------------------------------------*/

static int accumulate_neighbors8(SIPImage *image,int x0,int y0,int *hist){
	int	x,y;
	int	xs,xe,ys,ye;
	uchar	d,*p;
	xs = x0 -1;	xe = x0 +1;
	ys = y0 -1;	ye = y0 +1;
	d = ((uchar *)sip_getimgptr(image,y0))[x0];
	for ( y=ys ; y<=ye ; y++ ){
		p = (uchar *)sip_getimgptr(image,y);
		for ( x=xs ; x<=xe ; x++ ){
			if ( p[x] != d )  hist[ (uint)p[x] ]++;
		}
	}
	return(0);
}


static int make_inhibittable(SIPImage *image,char **inh_table,int maxlabelid){
	int	i,j,x,y,cid,nrid;
	int	width,height;
	int	nedgepel,npixel;
	int	*hist;
	double	adjacency;
	SIPImage	*distimage;
	GRPLST	ccpool;
	CRects	runpool(512);
	uchar	*p;
	uchar	d = 0;
	CRect	*rect;
	int	maxpos,maxval;

	width = image->width;
	height = image->height;
	if ( 0 == (hist = new int[ maxlabelid > 0 ? maxlabelid : 1 ]) ){
		return(-1);
	}
	if ( 0 == (distimage = sip_CreateImage(width,height,8)) ){
		delete []hist;
		return(-1);
	}
	sip_distimage(image,distimage,2);

	/* clear Connection-Inbition Table */
	for ( i=0 ; i<maxlabelid ; i++ ){
		for ( j=0 ; j<maxlabelid ; j++ )  inh_table[i][j] = 0;
	}

	/* create 8-connected components */
	xip_CreateCC(image,&ccpool,&runpool,1);
	for ( cid=0 ; cid < ccpool.groups ; cid++ ){
		if ( 0 >= (nrid = ccpool.getidcounts(cid)) )  continue;
		npixel = nedgepel = 0;
		for ( i=0 ; i<maxlabelid ; i++ )  hist[i] = 0;
		for ( i=0 ; i<nrid ; i++ ){
			rect = runpool.getrect( ccpool.getidlist(cid)[i] );
			y = rect->y1;
			if ( y == 0 || y == (height -1) )  continue;
			p = (uchar *)sip_getimgptr(distimage,y);
			npixel += rect->x2 - rect->x1 +1;
			for ( x = rect->x1 ; x <= rect->x2 ; x++ ){
				if ( p[x] != 1 )  continue;
				if ( x == 0 || x == (width -1) )  continue;
				if ( y == 0 || y == (height -1) )  continue;
				accumulate_neighbors8(image,x,rect->y1,hist);
				nedgepel++;
			}
		}
		maxval = maxpos = 0;
		for ( i=0 ; i<maxlabelid ; i++ ){
			if ( maxval >= hist[i] )  continue;
			maxval = hist[ maxpos = i ];
		}
		rect = runpool.getrect( ccpool.getidlist(cid)[0] );
		if ( nedgepel == 0 )  nedgepel++;
		adjacency = 100 * (double)maxval / (double)nedgepel /3;
//		if ( adjacency > 100 )  adjacency = 100;
#if 0
		printf("cid=%d (%d,%d), npixel=%d, nedgepel=%d (%.1f%%), ",cid,\
			rect->x1,rect->y1,npixel,nedgepel,\
			(100*(double)nedgepel/(double)npixel));
		printf("%d-%d (%.1f%%)\n",(uint)d,maxpos,adjacency);
#endif

#ifdef IgnoreFragmentSize
		if ( npixel <= IgnoreFragmentSize )  continue;
#endif

		if ( adjacency > AllowConflictEdgeRatio ){
			d = ((uchar *)sip_getimgptr(image,rect->y1))[rect->x1];
			inh_table[(uint)d][maxpos] = 1;
			inh_table[maxpos][(uint)d] = 1;
		}
	}
	ccpool.clearall();
	runpool.clear();
	sip_DestroyImage(distimage);
	delete []hist;
	return(0);
}


static int make_inhibittable_all(SIPImage *labelimage, \
		MTCELL **cmatrix,int mwidth,int mheight){
	int	x,y;
	int	mx,my;
	MTCELL	*cmatp;
	GLAB	*src;
	uchar	*dst;
	SIPImage	*label_local;

	if ( 0 == (label_local = sip_CreateImage(SUBREGIONSIZE,SUBREGIONSIZE,8)) ){
		return(-1);
	}

	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			cmatp->inh_table = alloc_inh_table(cmatp->ncolor);

			/* copy local label image */
			for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
				src = (GLAB *)sip_getimgptr(labelimage, \
					y + my * SUBREGIONSIZE);
				src += mx * SUBREGIONSIZE;
				dst = (uchar *)sip_getimgptr(label_local,y);
				for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
					dst[x] = (uchar)src[x];
				}
			}

			make_inhibittable(label_local,cmatp->inh_table,cmatp->ncolor);
		}
	}

	sip_DestroyImage(label_local);
	return(0);
}




/*----------------------------------------------
    Check conflict of global labels
----------------------------------------------*/

/* fast version  O(N) */
static int isconflict(LSETS *lsets,int *glmap,int *glrevtbl, \
		MTCELL **cmatrix,int gl1,int gl2,int maxcount){
	int	i,j;
	int	sgl1,sgl2;
	int	sid1,sid2;
	int	count;
	IDXSET	*idxset1,*idxset2;
	int	mx,my;
	MTCELL	*cmatp;
	                            /* set-ID */
	idxset1 = lsets->settbl[ sid1 = lsets->revtbl[gl1] ];
	idxset2 = lsets->settbl[ sid2 = lsets->revtbl[gl2] ];
	if ( idxset1 == idxset2 )  return(0);
	if ( idxset1->counts > idxset2->counts ){
		i = gl1;  gl1 = gl2;  gl2 = i;
		idxset1 = lsets->settbl[ sid1 = lsets->revtbl[gl1] ];
		idxset2 = lsets->settbl[ sid2 = lsets->revtbl[gl2] ];
	}
	for ( i=count=0 ; i<idxset1->counts ; i++ ){
		/* global-label id for scanning */
		sgl1 = idxset1->idlist[i];
		my = glmap[sgl1] / Max_mwidth;
		mx = glmap[sgl1] % Max_mwidth;
		cmatp = &cmatrix[my][mx];

		for ( j=0 ; j < cmatp->ncolor && count < maxcount ; j++ ){
			/* global-label id for scanning */
			sgl2 = cmatp->glabel[j];
				/* Skip if the global label is not in the set #2. */
			if ( sid2 != lsets->revtbl[sgl2] )  continue;
				/* Skip if the global label is in the same set. */
			if ( sid1 == lsets->revtbl[sgl2] )  continue;
			if ( cmatp->inh_table[ glrevtbl[sgl1] ][j] )  count++;
//			printf("%d,%d: %d %d\n",mx,my,glrevtbl[sgl1],j);
		}

		if ( count >= maxcount )  break;
	}
	return(count);
}




/*----------------------------------------------
    Merge labels
----------------------------------------------*/

static int mergelabel_2(LSETS *lsets,int *glmap,int *glrevtbl, \
		MTCELL **cmatrix,int mx1,int my1,int mx2,int my2, \
		int diff,int conflict){
	int	i,j,d;
	int	gl1,gl2;
	MTCELL	*cmatp1,*cmatp2;
	cmatp1 = &cmatrix[my1][mx1];
	cmatp2 = &cmatrix[my2][mx2];
	for ( i=0 ; i < cmatp1->ncolor ; i++ ){
		for ( j=0 ; j < cmatp2->ncolor ; j++ ){
			d = (int)cmatp1->collist[i].g - (int)cmatp2->collist[j].g;
			if ( d >= diff || d <= -diff )  continue;

			gl1 = cmatp1->glabel[i];  gl2 = cmatp2->glabel[j];

			if ( conflict != -1 ){
				if ( isconflict(lsets,glmap,glrevtbl, \
					cmatrix,gl1,gl2,\
					conflict +1) > conflict ){
					continue;
				}
			}

			if ( gl1 > gl2 ){
				gl2 = cmatp1->glabel[i];  gl1 = cmatp2->glabel[j];
			}
			if ( 0 > lsets->mergelabel(gl1,gl2) ){
				fprintf(stderr,"mergelabel() failed.\n");  return(-1);
			}
		}
	}
	return(0);
}


static int mergelabel(LSETS *lsets,int glcount,int *glmap,int *glrevtbl, \
		MTCELL **cmatrix,int mwidth,int mheight){
	int	i,id,sid,count;
	int	mx,my;
	IDXSET	*idxset;
	int	glevel;
	int	single_label[ (GrayLevels/TruncLevel) +1];

	/* At the begining, each global label should belong to a label set. */
	for ( i=0 ; i<glcount ; i++ ){
		if ( 0 > (id = lsets->newset() ) ){
			fprintf(stderr,"newset() failed.\n");  return(-1);
		}
		if ( 0 > lsets->addmember(id,i) ){
			fprintf(stderr,"addmember() failed.\n");  return(-1);
		}
	}

	/* merge labels among 8-neighbor cells */
	for ( i=0 ; mindiff[i]>=0 ; i++ ){
		for ( my=0 ; my<mheight ; my++ ){
			for ( mx=0 ; mx<mwidth ; mx++ ){
				if ( mx > 0 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx-1,my, \
						mindiff[i],\
						AllowConflict) )  return(-1);
				}
				if ( my > 0 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx,my-1, \
						mindiff[i], \
						AllowConflict) )  return(-1);
				}
				if ( mx > 0 && my > 0 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx-1,my-1, \
						mindiff[i>2 ? i-3:0], \
						AllowConflict) )  return(-1);
				}
				if ( mx < mwidth -1 && my > 0 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx+1,my-1, \
						mindiff[i>2 ? i-3:0], \
						AllowConflict) )  return(-1);
				}
			}
		}
	}

#ifdef	Merge5x5Mode
	/* merge labels among adjacent (+2 upper, +2 left) cells */
	for ( i=0 ; mindiff[i]>=0 ; i++ ){
		for ( my=0 ; my<mheight ; my++ ){
			for ( mx=0 ; mx<mwidth ; mx++ ){
				if ( mx > 1 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx-2,my, \
						mindiff[i], \
						AllowConflict) )  return(-1);
				}
				if ( my > 1 ){
					if ( mergelabel_2(lsets,glmap,glrevtbl, \
						cmatrix,mx,my,mx,my-2, \
						mindiff[i], \
						AllowConflict) )  return(-1);
				}
			}
		}
	}
#endif

#ifdef	TruncateTinyGroups
	/* truncate tiny groups */
	for ( i=0 ; i <= (GrayLevels/TruncLevel) ; i++ )  single_label[i] = -1;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			for ( i=0 ; i<cmatrix[my][mx].ncolor ; i++ ){
				sid = lsets->revtbl[ cmatrix[my][mx].glabel[i] ];
				idxset = lsets->settbl[sid];
				if ( idxset->counts >= TruncCounts )  continue;
				glevel = cmatrix[my][mx].collist[i].g / TruncLevel;
				if ( single_label[glevel] != -1 ){
 					lsets->mergelabel(single_label[glevel], \
						idxset->idlist[0]);
				}
				else{	single_label[glevel] = idxset->idlist[0];
				}
			}
		}
	}
#endif

	if ( sw_verbose ){
		for ( sid=count=0 ; sid<glcount ; sid++ ){
			if ( 0 != lsets->settbl[sid] )  count++;
		}
		printf("Number of labels after merging: %d\n",count);
	}

	return(0);
}




/*----------------------------------------------
    Assign global labels
----------------------------------------------*/

static int assign_globallabels(MTCELL **cmatrix, \
		int mwidth,int mheight,int **glrevtbl){
	int	mx,my,j;
	MTCELL	*cmatp;
	int	glcount;	/* global label counter */
	glcount = 0;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			for ( j=0 ; j<cmatp->ncolor ; j++ ){
				cmatp->glabel[j] = glcount++;
			}
		}
	}
	if ( glrevtbl == 0 )  return(glcount);
	if ( 0 == (*glrevtbl = new int[glcount > 0 ? glcount : 1]) )  return(-1);
	glcount = 0;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			for ( j=0 ; j<cmatp->ncolor ; j++ ){
				(*glrevtbl)[glcount++] = j;
			}
		}
	}
	return(glcount);
}




/*----------------------------------------------
    Make global label mapping table
        (global label -> subregion ID.)
----------------------------------------------*/

static int * mk_glmap(int glcount,MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my,i;
	MTCELL	*cmatp;
	int	*glmap;
	if ( 0 == (glmap = new int[glcount]) )  return(0);
	for ( i=0 ; i<glcount ; i++ )  glmap[i] = -1;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			for ( i=0 ; i<cmatp->ncolor ; i++ ){
				glmap[ cmatp->glabel[i] ] = my * Max_mwidth + mx;
			}
		}
	}
	return(glmap);
}




/*----------------------------------------------
    Make&save label image
----------------------------------------------*/

static int save_labelimage(SIPImage *labelimage,LSETS *lsets, \
		int glcount,int width,int height,MTCELL **cmatrix, \
		int mwidth,int mheight,char *fname){
	SIPImage	*simage;
	GLAB	*src;
	uchar	*dst;
	int	y,mx,my;
	int	sw,sh;
	int	sid;
	uint	count;
	IDXSET	*idxset;
	PBMFILE	PBMSV;
	uchar	*xlat;
	register int	x;
	register int	*glabel;
	int	*revtbl;
	MTCELL	*cmatp;
	if ( 0 == (simage = sip_CreateImage(width,height,8)) )  return(-1);
	if ( 0 == (xlat = new uchar[ glcount > 0 ? glcount : 1]) ){
		sip_DestroyImage(simage);  return(-1);
	}

	for ( sid=count=0 ; sid<glcount ; sid++ ){
		if ( 0 == (idxset = lsets->settbl[sid]) )  xlat[sid] = 0;
		else  xlat[sid] = (uchar)count++;
	}

	revtbl = lsets->revtbl;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			glabel = cmatp->glabel;
			sw = sh = SUBREGIONSIZE;
			if ( sw + mx * SUBREGIONSIZE > width ){
				sw = width - mx * SUBREGIONSIZE;
			}
			if ( sh + my * SUBREGIONSIZE > height ){
				sh = height - my * SUBREGIONSIZE;
			}
			for ( y=0 ; y<sh ; y++ ){
				src = (GLAB *)sip_getimgptr(labelimage,\
					y + my * SUBREGIONSIZE);
				src += mx * SUBREGIONSIZE;
				dst = (uchar *)sip_getimgptr(simage,\
					y + my * SUBREGIONSIZE);
				dst += mx * SUBREGIONSIZE;
				for ( x=0 ; x<sw ; x++ ){
					dst[x] = xlat[ revtbl[ glabel[(int)src[x]] ] ];
				}
			}
		}
	}

	xip_SaveImage(fname,simage);

	delete []xlat;
	sip_DestroyImage(simage);
	return(count);
}




/*----------------------------------------------
    Make&save split images
----------------------------------------------*/

static int save_splitimages(SIPImage *labelimage,LSETS *lsets, \
		int glcount,int width,int height,MTCELL **cmatrix, \
		int mwidth,int mheight){
	SIPImage	*simage;
	GLAB	*src;
	uchar	*dst;
	int	y,mx,my;
	int	sw,sh;
	int	sid,count,retcode;
	char	outfile[30];
	IDXSET	*idxset;
	PBMFILE	PBMSV;
	int	x,npixel;
	int	*glabel;
	int	*revtbl;
	MTCELL	*cmatp;
	if ( 0 == (simage = sip_CreateImage(width,height,8)) )  return(-1);
	
	revtbl = lsets->revtbl;
	for ( sid=count=0 ; sid<glcount ; sid++ ){
		if ( 0 == (idxset = lsets->settbl[sid]) )  continue;

#ifdef	SaveFilesLimit
		if ( count +1 >= SaveFilesLimit ){
			break;
		}
#endif

//		sip_ClearImage(simage,1);
		npixel = 0;
		for ( my=0 ; my<mheight ; my++ ){
			for ( mx=0 ; mx<mwidth ; mx++ ){
				cmatp = &cmatrix[my][mx];
				glabel = cmatp->glabel;
				sw = sh = SUBREGIONSIZE;
				if ( sw + mx * SUBREGIONSIZE > width ){
					sw = width - mx * SUBREGIONSIZE;
				}
				if ( sh + my * SUBREGIONSIZE > height ){
					sh = height - my * SUBREGIONSIZE;
				}
				for ( y=0 ; y<sh ; y++ ){
					src = (GLAB *)sip_getimgptr(labelimage,\
						y + my * SUBREGIONSIZE);
					src += mx * SUBREGIONSIZE;
					dst = (uchar *)sip_getimgptr(simage,\
						y + my * SUBREGIONSIZE);
					dst += mx * SUBREGIONSIZE;
					for ( x=0 ; x<sw ; x++ ){
						if ( revtbl[ glabel[(int)src[x]] ] == sid ){
							dst[x] = 0;
							++npixel;
						}
						else{	dst[x] = 1;
						}
					}
				}
			}
		}

		/* If the number of black pixel is less than Min_ObjectSize, */
		/* skip saving the split image. */
		if ( npixel < Min_ObjectSize )  continue;

		sprintf(outfile,"%04d.pbm",count);
		PBMSV.setsize(width,height,1);
		if ( 0 > PBMSV.open(outfile,"w") ){
			count = -1;  break;
		}
		retcode = 0;
		for ( y=0 ; y<height ; y++ ){
			if ( PBMSV.writeline_bilevel(-1, \
				(void *)sip_getimgptr(simage,y),1) ){
				retcode = -1;  break;
			}
		}
		if ( retcode ) PBMSV.close();  else retcode = PBMSV.close();
		if ( retcode ){
			count = -1;  break;
		}

		count++;
	}

	sip_DestroyImage(simage);
	return(count);
}




/*----------------------------------------------
    Show average number of local grays/colors
----------------------------------------------*/

static void show_ncol_ave(MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my;
	int	ave = 0;
	for ( mx=my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			ave += cmatrix[my][mx].ncolor;
		}
	}
	printf("Average number of local grays(colors) = %.3f\n",\
		(double)ave/(double)(mx * my));
}




/*----------------------------------------------
    Load PBM/PGM/PPM image file
----------------------------------------------*/

static int load_image(char *fname,SIPImage **image, \
		int *width,int *height,int bufdepth){
	PBMFILE	PBMLD;
	int	retcode,depth,x,y;
	SIPImage	*imgbuf;
	uchar	*lbuf,*src,*dst;
	uint	c;
	int	bufwidth,bufheight;
	*image = 0;	/* default value = NULL */
	PBMLD.setpal_fb(0,0xff);
	if ( bufdepth != 8 && bufdepth != 32 )  return(-5);
	retcode = PBMLD.open(fname,"r");
	if ( retcode )  return(retcode);
	switch ( PBMLD.getfiletype() ){
	  case 0:	depth = 1;  break;
	  case 1:	depth = 8;  break;
	  case 2:	depth = 32;  break;	/* 24bit */
	  default:	return(-5);
	}
	*width  = PBMLD.getwidth();
	*height = PBMLD.getheight();

	bufwidth = ((*width + SUBREGIONSIZE -1) / SUBREGIONSIZE) * SUBREGIONSIZE;
	bufheight = ((*height + SUBREGIONSIZE -1) / SUBREGIONSIZE) * SUBREGIONSIZE;

	if ( 0 == (lbuf = new uchar[3 * bufwidth]) )  return(-6);
	if ( 0 == (imgbuf = sip_CreateImage(bufwidth,bufheight,bufdepth)) ){
		delete []lbuf;
		return(-6);
	}
	sip_ClearImage(imgbuf,0xff);
	retcode = 0;
	for ( y=0 ; y < *height ; y++ ){
		if ( depth == 1 ){
			if ( 0 != PBMLD.readline_gray(-1,(void *)lbuf) ){
				retcode = -1;	break;	/* read error */
			}
		}
		else{	/* depth=8, 32 */
			if ( 0 != PBMLD.readline(-1,(void *)lbuf) ){
				retcode = -1;	break;	/* read error */
			}
		}
		if ( 32 == depth ){
			src = lbuf;  dst = (uchar *)sip_getimgptr(imgbuf,y);
			if ( bufdepth == 8 ){
				for ( x=0 ; x < *width ; x++ ){
					c = (uint)(.299 * (double)((uint)src[0]) \
						 + .587 * (double)((uint)src[1]) \
						 + .114 * (double)((uint)src[2]));
					if ( c > 255 )  c = 255;
					*dst++ = c;
					src += 3;
				}
			}
			else{	for ( x=0 ; x < *width ; x++ ){
					dst[0] = src[0];
					dst[1] = src[1];
					dst[2] = src[2];
					dst[3] = 0;
					src += 3;  dst += 4;
				}
			}
		}
		else{	if ( bufdepth == 8 ){
				memcpy((void *)sip_getimgptr(imgbuf,y), \
					(void *)lbuf,*width);
			}
			else{	src = lbuf;
				dst = (uchar *)sip_getimgptr(imgbuf,y);
				for ( x=0 ; x < *width ; x++ ){
					dst[0] = dst[1] = dst[2] = *src++;
					dst[3] = 0;
					dst += 4;
				}
			}
		}
	}
	delete []lbuf;
	PBMLD.close();
	if ( ! retcode ){
		*image = imgbuf;
	}
	else{	sip_DestroyImage(imgbuf);
	}
	return(retcode);
}




/*----------------------------------------------
    Print usage
----------------------------------------------*/

static void usage(){
	fputs("dlabel v1.31 Copyright (C) 1999-2002 Hideaki Goto \n",stderr);
	fputs("Usage: dlabel [-options] in_file \n",stderr);
	fputs("         -s         :  smooth input image (k-nearest neighbor method) \n",stderr);
	fputs("         -w         :  use weighted-histogram \n",stderr);
//	fputs("         -edge      :  edge compensation \n",stderr);
	fputs("         -ri file   :  create reduced image \n",stderr);
	fputs("         -label file:  create label image \n",stderr);
	fputs("         -decomp    :  create decomposed images (saved as ./XXXX.pbm) \n",stderr);
	fputs("         -verbose   :  show processing information \n",stderr);
}




/*----------------------------------------------
    Main routine
----------------------------------------------*/

int main(int ac,char *av[]){
	char	*infile  = NULL;
	char	*rimagefile = 0;
	char	*limagefile = 0;
	int	width,height,mwidth,mheight;
	int	i,k;
	SIPImage	*image;
	SIPImage	*rimage = 0;
	SIPImage	*labelimage;
	MTCELL	**cmatrix;
	int	glcount;
	int	*glmap;		/* global label -> subregion translation table */
	int	*glrevtbl;	/* global label -> local label translation table */
	LSETS	LSETS;

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-ri",av[i]) ){
			if ( i++ >= ac ){  k=0; continue; }
			rimagefile = av[i];
			continue;
		}
		if ( 0 == strcmp("-label",av[i]) ){
			if ( i++ >= ac ){  k=0; continue; }
			limagefile = av[i];
			continue;
		}
		if ( 0 == strcmp("-s",av[i]) ){  sw_smooth = -1;  continue; }
		if ( 0 == strcmp("-w",av[i]) ){  sw_weighted = -1;  continue; }
		if ( 0 == strcmp("-edge",av[i]) ){  sw_edgecomp = -1;  continue; }
		if ( 0 == strcmp("-decomp",av[i]) ){  sw_decomp = -1;  continue; }
		if ( 0 == strcmp("-verbose",av[i]) ){  sw_verbose = -1;  continue; }
		if ( 0 == strcmp("-v",av[i]) ){  sw_verbose = -1;  continue; }
		if ( av[i][0] == '-' ){  k=0;  continue; }
		if ( NULL == infile ){  infile = av[i];  continue; }
	}
	if ( k == 0 || infile == NULL ){  usage();  return(0); }


/* ---- Read Image ---- */

	time_start();
	if ( 0 > load_image(infile,&image,&width,&height,IMGBUFDEPTH) ){
		fprintf(stderr,"Can't read \"%s\".\n",infile);
		return(2);
	}
	time_get(&utime_load,&stime_load);


/* ---- Allocate Buffers ---- */

	mwidth  = image->width  / SUBREGIONSIZE;
	mheight = image->height / SUBREGIONSIZE;
	if ( sw_verbose ){
		printf("Image  size = %d x %d\n",width,height);
		printf("Buffer size = %d x %d\n",image->width,image->height);
		printf("Matrix size = %d x %d\n",mwidth,mheight);
	}

	if ( rimagefile ){
		if ( 0 == (rimage = sip_CreateImage(image->width,image->height,IMGBUFDEPTH)) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
	}

	if ( 0 == (labelimage = sip_CreateImage(image->width, \
				image->height,sizeof(GLAB) * 8)) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}

	if ( 0 == (cmatrix = new MTCELL*[ mheight ]) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}
	for ( i=0 ; i<mheight ; i++ ){
		if ( 0 == (cmatrix[i] = new MTCELL[ mwidth ]) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
	}


/* ---- Image Filtering (simple k-NN smoothing) ---- */

	utime_filt = stime_filt = 0;

	if ( sw_smooth ){
		time_start();
		sip_kNNsmooth(image,image,3,0);
		if ( sw_verbose )  printf("Image filtering is done.\n");
		time_get(&utime_filt,&stime_filt);
	}


/* ---- Create Gray Level Histograms ---- */

	utime_edge = stime_edge = 0;
	if ( sw_weighted ){
		time_start();
		if ( 0 > findgray_pd_setweight(image) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
		if ( sw_verbose )  printf("Edge detection is done.\n");
		time_get(&utime_edge,&stime_edge);
	}


/* ---- Local Image Labeling ---- */

	time_start();
	find_graylevels(image,cmatrix,mwidth,mheight,sw_weighted);

	make_locallabelimage(image,labelimage,cmatrix,mwidth,mheight);

	glcount = assign_globallabels(cmatrix,mwidth,mheight,&glrevtbl);
	if ( glcount < 0 ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}
	if ( sw_verbose ){
		printf("Gray level / Color detection & image labeling are done.\n");
		show_ncol_ave(cmatrix,mwidth,mheight);
	}
	time_get(&utime_label,&stime_label);


/* ---- Merge Labels ---- */

	time_start();
	make_inhibittable_all(labelimage,cmatrix,mwidth,mheight);
	if ( 0 == (glmap = mk_glmap(glcount,cmatrix,mwidth,mheight)) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}
	mergelabel(&LSETS,glcount,glmap,glrevtbl,cmatrix,mwidth,mheight);
	time_get(&utime_merge,&stime_merge);


/* ---- Make Color-reduced Image ---- */

	utime_reduce = stime_reduce = 0;
	if ( rimage ){
		time_start();
		make_reducedimage(labelimage,rimage,cmatrix,mwidth,mheight);
		xip_SaveImage(rimagefile,rimage);
		time_get(&utime_reduce,&stime_reduce);
	}


/* ---- Save Label Image ---- */

	utime_mklab = stime_mklab = 0;
	if ( limagefile ){
		time_start();
		if ( 0 > (i = save_labelimage(labelimage,&LSETS,glcount,width,height, \
				cmatrix,mwidth,mheight,limagefile)) ){
			fputs("Can't write output file.\n",stderr);
			return(2);
		}
		time_get(&utime_mklab,&stime_mklab);
	}


/* ---- Save Decomposed Images ---- */

	utime_decomp = stime_decomp = 0;
	if ( sw_decomp ){
		time_start();
		if ( 0 > (i = save_splitimages(labelimage,&LSETS,glcount,width,height, \
				cmatrix,mwidth,mheight)) ){
		fputs("Can't write output file.\n",stderr);
		return(2);
	}
	time_get(&utime_decomp,&stime_decomp);
	}


/* ---- Print Some Information  ---- */

	if ( sw_verbose ){
		printf("Saved images: %d\n",i);
	}

	if ( sw_verbose ){
		printf("Time: Image_loading	%.3f(usr) %.3f(sys)\n", \
						utime_load,stime_load);
		if ( sw_smooth )
		  printf("Time: Image_filtering	%.3f(usr) %.3f(sys)\n", \
						utime_filt,stime_filt);
		if ( sw_weighted )
		  printf("Time: Edge_detection	%.3f(usr) %.3f(sys)\n", \
						utime_edge,stime_edge);
		printf("Time: Image_labeling	%.3f(usr) %.3f(sys)\n", \
						utime_label,stime_label);
		printf("Time: Label_merge	%.3f(usr) %.3f(sys)\n", \
						utime_merge,stime_merge);
		if ( rimage )
		  printf("Time: Mk_reduced_image	%.3f(usr) %.3f(sys)\n", \
						utime_reduce,stime_reduce);
		if ( limagefile )
		  printf("Time: Mk_label_image	%.3f(usr) %.3f(sys)\n", \
						utime_mklab,stime_mklab);
		if ( sw_decomp )
		  printf("Time: Mk_decomp_images	%.3f(usr) %.3f(sys)\n", \
						utime_decomp,stime_decomp);
		printf("Time: Total		%.3f(usr) %.3f(sys)\n", \
			(float)(utime_load + utime_filt + utime_edge \
			+ utime_label + utime_merge + utime_reduce \
			+ utime_mklab + utime_decomp), \
			(float)(stime_load + stime_filt + stime_edge \
			+ stime_label + stime_merge + stime_reduce \
			+ stime_mklab + stime_decomp) );
	}

	sip_DestroyImage(labelimage);
	if ( rimage )  sip_DestroyImage(rimage);
	sip_DestroyImage(image);
	return(0);
}


